       integer function lusolve(a,m,b,n,piv,atau,rtau)
!
!  NOTE: When piv present, a will be assumed to be an LU 
!          decomposition, with piv containing the pivot information
!
!        No piv in arglist: a will be decomposed, and rewritten 
!           as a LU decomposition
!
!
       implicit none

       integer m,n,i,j
       integer, optional :: piv(2,m-1)
       integer, allocatable :: pv(:,:)
       real a(m,m),b(m,n),rrtau,ax
       real, optional :: atau,rtau
       interface
        subroutine xrow(a,r1,r2)
         integer r1,r2
         real a(:,:)
        end subroutine xrow
        subroutine ludc(a,piv,m)
         integer m,piv(2,m-1)
         real a(m,m)
        end subroutine ludc
       end interface

       rrtau = 1.221E-4  !  0.5*sqrt(slamch('epsilon'))
       if( present(rtau) )then
        rrtau = rtau
       endif
       allocate(pv(2,m-1))

! when piv is present, a has been decomposed already 
       if( present(piv) )then
         pv(1:2,1:m-1) = piv(1:2,1:m-1)
       else
         call ludc(a,pv,m)
       endif

       do i = 1,m-1   !  apply the row pivots to b, forward order
        call xrow(b,i,pv(1,i))
       enddo

       do i = 2,m  !   solve the L triangle
        b(i:i,1:n) = b(i:i,1:n) - matmul(a(i:i,1:i-1),b(1:i-1,1:n))
       enddo

       lusolve = m
       do i = m,1,-1  !  solve the U triangle
        if( abs(a(i,i)) > tau() )then
         lusolve = lusolve - 1
         ax = 1./a(i,i)
         if( i == m )then
          b(i,1:n) = ax*b(i,1:n)
         else
          b(i:i,1:n) = ax*( b(i:i,1:n) - matmul(a(i:i,i+1:m),b(i+1:m,1:n)))
         endif
        endif
       enddo

       do i = m-1,1,-1   ! apply column pivots to rows, backward order
        call xrow(b,i,pv(2,i))
       enddo

       deallocate(pv)

       return

      contains

      real function tau()
       if( present(atau) )then
        tau = atau
       else
        if( i > 1 )then
         tau = rrtau*sqrt(sum(a(1:i-1,i)**2)/(i-1))
        else
         tau = 0.
        endif
       endif
       return
      end function tau

       end function lusolve
