        module smparith
!  signed integer arithmetic, up to 8192 digits in length
        implicit none

        type smprec
          integer s  !  sign
          integer l  ! used string length
          integer x  ! maximum string length
          character(len=1), pointer :: m  
        end type smprec

        logical smp_strip,smp_initiallized

        character(len=1),target :: zero_smp_c,one_smp_c
        character(len=1),target:: plusfcn_c(8192),multfcn_c(8192),&
        &dvdefcn_c(8192),sbtrfcn_c(8192)
        type( smprec )  zero_smp,one_smp
        integer smparith_ii1,smparith_ii2,smp_nhlast,smp_trignh,&
        & smparith_macc
        double complex, allocatable :: smp_trig(:)
        save smparith_ii1,smparith_ii2,zero_smp,one_smp,smp_strip,zero_smp_c,&
        &one_smp_c,plusfcn_c,multfcn_c,dvdefcn_c,sbtrfcn_c,smp_trig,smp_nhlast,&
        &smp_trignh
        data smp_nhlast/0/,smp_trignh/0/,smp_initiallized/.false./,&
        & smparith_macc/0/

       interface
         subroutine fht_dp(x,trig,n)
          real(8) x(*)
          double complex trig(*)
          integer, value :: n
         end subroutine fht_dp
          subroutine trigs_dp(trig,n)
           double complex trig(*)
           integer, value :: n
          end subroutine trigs_dp
       end interface

        interface assignment( = )
         module procedure set_smprec
        end interface

        interface operator( + )
         module procedure plus_smprec
        end interface

        interface operator( - )
         module procedure sbtr_smprec
        end interface

        interface operator( * )
         module procedure mult_smprec
        end interface

        interface operator( / )
         module procedure dvde_smprec
        end interface

         interface operator( .gt. )
           module procedure gt_smprec
         end interface
 
         interface operator( .lt. )
           module procedure lt_smprec
         end interface
 
         interface operator( .ge. )
           module procedure ge_smprec
         end interface
 
        interface operator( .eq. )
          module procedure eq_smprec
        end interface
 
        interface operator( .ne. )
          module procedure ne_smprec
        end interface
 
         interface operator( .le. )
           module procedure le_smprec
         end interface

         interface operator( .mgt. )
           module procedure mgt_smprec
         end interface
 
         interface operator( .mlt. )
           module procedure mlt_smprec
         end interface

         interface operator( .mge. )
           module procedure mge_smprec
         end interface

         interface operator( .meq. )
           module procedure meq_smprec
         end interface
 
         interface operator( .mne. )
           module procedure mne_smprec
         end interface

         interface operator( .mle. )
           module procedure mle_smprec
         end interface
 
        contains

      subroutine smp_trigs(nh)
      integer nh
      if( .not. allocated(smp_trig) )then
       allocate(smp_trig(1:nh))
       smp_nhlast = nh
       call trigs_dp(smp_trig,nh)
       smp_trignh = nh
      elseif( nh /= smp_trignh )then
       if( nh > smp_nhlast )then
        deallocate(smp_trig)
        allocate(smp_trig(1:nh))
        smp_nhlast = nh
       endif
       call trigs_dp(smp_trig,nh)
       smp_trignh = nh
      endif
      return
      end subroutine smp_trigs

      subroutine set_smprec(u,v)
      type(smprec),intent(out) :: u
      type(smprec),intent(in) :: v
!d      integer i
!  u%x and v%x were assigned when they were initiallized;
!  these do not change; the pointers stay the same as well
!      write(*,*)' setsmp ',v%l
      if( u%x < v%l )then
       write(0,'(1x,a)')'set_smprec error: x <= l'
       call exit(1)
      else
       u%s = v%s
       u%l = v%l
       u%m(1:v%l) = v%m(1:v%l)
      endif
!d      write(*,*)' return setsmp ',u%l,v%l
!d      write(*,*)(ichar(u%m(i:i)),i=1,u%l)
      return
      end subroutine set_smprec

      subroutine smpinit
        integer j
        character*1 creg(4)
        character*4 cireg
        equivalence (cireg,creg)
        smp_initiallized = .true.
        cireg=transfer(256*ichar('2')+ichar('1'),cireg)
        do  j=1,4
          if (creg(j).eq.'1') smparith_ii1=j
          if (creg(j).eq.'2') smparith_ii2=j
        enddo
        zero_smp%m => zero_smp_c
        zero_smp%x = 1
        zero_smp%l = 1  
        zero_smp%s = 1
        zero_smp%m(1:1) = char(0)
        one_smp%m => one_smp_c
        one_smp%x = 1
        one_smp%l = 1
        one_smp%s = 1
        one_smp%m(1:1) = char(1)
        smp_strip = .true.
        smparith_macc=29
      return
      end subroutine smpinit

      subroutine smpadd(w,ui,vi)
      integer loc1,j
      character(len=1),target:: u_c(8192),v_c(8192)
      type(smprec) :: w,u,v,ui,vi
      character*1 creg(4)
      character*4 cireg
      equivalence (cireg,creg)
       cireg=transfer(0,cireg)
!       write(*,*)' smpadd ui%l ',ui%l
!       call display(ui,'ui =')
!       write(*,*)' smpadd vi%l ',vi%l
!       call display(vi,'vi =')
!  have to do this because w maybe one of ui or vi
       u%m => u_c(1)
       v%m => v_c(1)
       u%x = 8192
       v%x = 8192
       u = ui 
       v = vi  
!       call display(u,'smpadd u')
!       call display(v,'smpadd v')
       loc1 = max(u%l,v%l)   !  line up the units place
!       write(*,*)'u%l,v%l,loc1',u%l,v%l,loc1
       call smprshft(v,loc1-v%l)  
       call smprshft(u,loc1-u%l)

       if ( u%s .eq. v%s )then  !  u + , v +    .or. u - , v - 
        if( u .mge. v )then
!        write(*,*)' same sign, u mge v loop'
         do  j=u%l,1,-1
           cireg=transfer(ichar(u%m(j:j))+ichar(v%m(j:j))+ichar(creg(smparith_ii2)),cireg)
           w%m(j+1:j+1)=creg(smparith_ii1)
         enddo
         w%m(1:1)=creg(smparith_ii2)
         w%s = u%s
         w%l = u%l + 1
        else  !  u .mlt. v
!        write(*,*)' same sign, u mlt v loop'
         do  j=v%l,1,-1
           cireg=transfer(ichar(v%m(j:j))+ichar(u%m(j:j))+ichar(creg(smparith_ii2)),cireg)
           w%m(j+1:j+1)=creg(smparith_ii1)
         enddo
         w%m(1:1)=creg(smparith_ii2)
         w%s = v%s
         w%l = v%l + 1
        endif
       elseif( u%s .eq. 1 )then  ! u + , v -
        if( u .mge. v )then
!         write(*,*)' u+v-, u mge v loop'
         cireg=transfer(256,cireg) 
         do  j=u%l,1,-1
           cireg=transfer(255+ichar(u%m(j:j))-ichar(v%m(j:j))+ichar(creg(smparith_ii2)),cireg)
           w%m(j:j)=creg(smparith_ii1)
         enddo
         w%s = u%s
         w%l = u%l
        else !  v .mgt. u 
!         write(*,*)' u+v-, u mlt v loop'
         cireg=transfer(256,cireg) 
         do  j=v%l,1,-1
           cireg=transfer(255+ichar(v%m(j:j))-ichar(u%m(j:j))+ichar(creg(smparith_ii2)),cireg)
           w%m(j:j)=creg(smparith_ii1)
         enddo
         w%s = v%s
         w%l = v%l
        endif
       else  ! u - , v +
        if( v .mge. u )then
!         write(*,*)' u-v+, u mle v loop'
         cireg=transfer(256,cireg) 
         do  j=v%l,1,-1
           cireg=transfer(255+ichar(v%m(j:j))-ichar(u%m(j:j))+ichar(creg(smparith_ii2)),cireg)
           w%m(j:j)=creg(smparith_ii1)
         enddo
         w%s = v%s
         w%l = v%l
        else ! u .mgt. v
!         write(*,*)' u-v+, u mgt v loop'
         cireg=transfer(256,cireg) 
         do  j=u%l,1,-1
           cireg=transfer(255+ichar(u%m(j:j))-ichar(v%m(j:j))+ichar(creg(smparith_ii2)),cireg)
           w%m(j:j)=creg(smparith_ii1)
         enddo
         w%s = u%s
         w%l = u%l
        endif
       endif  !  end u%s,v%s branches
       if( smp_strip )then
        call smpzstrip(w,0)
       endif
      return
      end subroutine smpadd

      subroutine smpzstrip(w,nlv)
      type(smprec) w
      integer j,nlv
!   requires nlv >= 0
      j = 0
      do while ( j+1 .lt. w%l .and. w%m(j+1:j+1) .eq. char(0) )
       j = j + 1
      enddo
      if( j .gt. 0 )then
       call smplshft(w,j-nlv)
      elseif( nlv .gt. 0)then
       call smprshft(w,nlv)
      endif
      return
      end subroutine smpzstrip


      subroutine smpssub(is,w,u,v)
      type(smprec) :: w,u,v
      integer j,is
      character*1 creg(4)
      character*4 cireg
      equivalence (cireg,creg)
        cireg=transfer(256,cireg)
        do  j=u%l,1,-1
         cireg=transfer(255+ichar(u%m(j:j))-ichar(v%m(j:j))+ichar(creg(smparith_ii2)),cireg)
         w%m(j:j)=creg(smparith_ii1)
        enddo
        w%l = u%l
        w%s = u%s
        is=ichar(creg(smparith_ii2))-1
        return
      end subroutine smpssub

      subroutine smpsub(w,u,vi)
      character(len=1),target::v_c(8192)
      type(smprec) :: w,u,vi,v
      character*1 creg(4)
      character*4 cireg
      equivalence (cireg,creg)
      v%m=>v_c(1)
      v%x=8192
      v = vi
      v%s = -v%s  !  flip the sign of v
      call smpadd(w,u,v)
      return
      end subroutine smpsub



      subroutine smpsad(w,ui,iv)
      type(smprec) :: w,ui,u
      integer j,iv
      character*1 creg(4)
      character*4 cireg
      equivalence (cireg,creg)
      character(len=1),target::u_c(8192)
      u%m=>u_c(1)
      u%x=8192
        cireg=transfer(256*iv,cireg)
        u = ui
        w%l = u%l + 1
        do  j=u%l,1,-1
          cireg=transfer(ichar(u%m(j:j))+ichar(creg(smparith_ii2)),cireg)
          w%m(j+1:j+1)=creg(smparith_ii1)
        enddo
        w%m(1:1)=creg(smparith_ii2)
        w%s = u%s
        return
      end subroutine smpsad

      subroutine smpsmu(w,ui,iv)
      type(smprec) :: w,ui,u
      integer j,iv
      character*1 creg(4)
      character*4 cireg
      equivalence (cireg,creg)
      character(len=1),target::u_c(8192)
      u%m=>u_c(1)
      u%x=8192
        cireg=transfer(0,cireg)
        u = ui
        w%l = u%l + 1
        do  j=u%l,1,-1
          cireg=transfer(ichar(u%m(j:j))*iv+ichar(creg(smparith_ii2)),cireg)
          w%m(j+1:j+1)=creg(smparith_ii1)
        enddo
        w%m(1:1)=creg(smparith_ii2)
        w%s = u%s
        return
      end subroutine smpsmu

      subroutine smpsdv(w,ui,iv,ir)
      character(len=1),target:: u_c(8192)
      integer j,iv,ir,i
      type(smprec) :: w,ui,u
      character*1 creg(4)
      character*4 cireg
      equivalence (cireg,creg)
      u%m => u_c(1)
      u%x = 8192
        u = ui
        ir=0
        do  j=1,u%l
          i=256*ir+ichar(u%m(j:j))
          w%m(j:j)=char(i/iv)
          ir=mod(i,iv)
        enddo
        w%l = u%l
        w%s = u%s
        return
      end subroutine smpsdv

      subroutine smpsneg(u)
      type(smprec) :: u
      integer j
      character*1 creg(4)
      character*4 cireg
      equivalence (cireg,creg)
        cireg=transfer(256,cireg)
        do  j=u%l,1,-1
          cireg=transfer(255-ichar(u%m(j:j))+ichar(creg(smparith_ii2)),cireg)
          u%m(j:j)=creg(smparith_ii1)
        enddo
        return
      end subroutine smpsneg


      subroutine smplshft(u,nshift)
      type(smprec) :: u
      integer j,nshift
      character*1 creg(4)
      character*4 cireg
      equivalence (cireg,creg)
        if( nshift .le. 0 )then
          return
        elseif( nshift .gt. u%l )then
          u = zero_smp
          return
        endif
        u%l = u%l - nshift
        do j=1,u%l
          u%m(j:j)=u%m(j+nshift:j+nshift)
        enddo
        return
      end subroutine smplshft

      subroutine smprshft(u,nshift)
      type(smprec) :: u
      integer j,nshift
      character*1 creg(4)
      character*4 cireg
      equivalence (cireg,creg)
!       write(*,*)' into smprshft with ',(ichar(u%m(i:i)),i=1,u%l)
       if( nshift .le. 0 )then
        return
       endif
        u%l = u%l + nshift
        do j=u%l,nshift+1,-1
          u%m(j:j)=u%m(j-nshift:j-nshift)
        enddo
        u%m(1:nshift) = repeat(char(0),nshift)
!        write(*,*)' out of smprshft with ',(ichar(u%m(i:i)),i=1,u%l)
        return
      end subroutine smprshft

      subroutine smpdiv(q,r,ui,vi) 
      integer nmax,i,is
      type(smprec)  u,v,q,r,rr,s,ui,vi
      parameter (nmax=8192)
!u    uses smpinv,smpmul,smpssub
      character(len=1),target:: u_c(8192),v_c(8192),rr_c(8192),s_c(8192)
      u%m => u_c(1)
      u%x = 8192
      v%m => v_c(1)
      v%x = 8192
      rr%m => rr_c(1)
      rr%x = 8192
      s%m => s_c(1)
      s%x = 8192


1001  if( vi .meq. one_smp )then
       q = ui
       r = zero_smp
       if( vi%s < 0 )then
        q%s = -q%s
       endif
       return
      elseif( vi .meq. ui )then
       q = one_smp
       r = zero_smp
       if( vi%s /= ui%s )then
        q%s = -q%s
       endif
       return
      elseif( vi .mgt. ui )then
       q = zero_smp
       r = ui
       return
      elseif( vi .meq. zero_smp )then
       write(0,'(1x,a)')'error in smpdiv: v = 0'
       call exit(1)
      endif

      v = vi
      call smpzstrip(v,0)  !  used to be (v,1)
      smp_strip = .false.

      u = ui
      r%l = v%l
      q%l = u%l - v%l + 1
      if(u%l+smparith_macc.gt.nmax)then
        write(0,*)'u ',(ichar(u%m(i:i)),i=1,u%l)
        write(0,*)'v ',(ichar(v%m(i:i)),i=1,v%l)
        write(0,*)'nmax too small in smpdiv'
        call exit(1)
      endif
!      call display(u,'entering smpdiv with u=')
!      write(*,*)'u%l',u%l
!      write(*,*)(ichar(u%m(i:i)),i=1,u%l)
!      call display(v,'and v=')
!      write(*,*)'v%l',v%l
!      write(*,*)(ichar(v%m(i:i)),i=1,v%l)
      call smpinv(s,v,u%l-v%l+smparith_macc)
!      call display(s,'vinv')
!      write(*,*)'s%l',s%l
!      write(*,*)(ichar(s%m(i:i)),i=1,s%l)
      call smpmul(rr,s,u)
!      call display(rr,'vinv*u')
!      write(*,*)'old rr%l',rr%l
!      write(*,*)(ichar(rr%m(i:i)),i=1,rr%l)
      rr%l = u%l - v%l + 2
      call smplshft(rr,1)
!      call display(rr,'setting q =')
      q = rr
      call smpmul(rr,q,v)
!      call display(rr,'q*v')
!      write(*,*)'q*v rr%l',rr%l
!      write(*,*)(ichar(rr%m(i:i)),i=1,rr%l)
      call smplshft(rr,1)
!      call display(rr,'lsfht 1 q*v = rr =')
!      write(*,*)'lshft 1 rr%l',rr%l
!      write(*,*)(ichar(rr%m(i:i)),i=1,rr%l)
      call smpssub(is,rr,u,rr)
      if(is .ne. 0)then
       smparith_macc = smparith_macc + 11
       if( smparith_macc > 4000 )then
        write(0,*)'smparith_macc > 4000 in smpdiv'
        call exit(1)
       endif
       go to 1001
      endif
      r = rr
!      call display(r,'r after r=rr')
!      write(*,*)'r%l',r%l
!      write(*,*)(ichar(r%m(i:i)),i=1,r%l)
      call smplshft(r,u%l-v%l)
!      call display(r,'lsfht u%l-v%l r')
!      write(*,*)'r%l',r%l
!      write(*,*)(ichar(r%m(i:i)),i=1,r%l)
      if( r .mge. v )then   !  sometimes the inverse doesn't quite do it !!!!
!       write(*,*)'|r| too large; u%s,v%s',u%s,v%s
!       call display(q,'before q')
!       call display(r,'before r')
       if( u%s*v%s > 0 )then
         q = q + one_smp
         r = r - v
       else
         q = q - one_smp
         r = r + v
       endif
!       call display(q,'after q')
!       call display(r,'after r')
      endif
      q%s = u%s*v%s
      r%s = u%s
      smp_strip = .true.
      call smpzstrip(q,0)
      call smpzstrip(r,0)
      return
      end subroutine smpdiv


      subroutine smpmul(w,ui,vi) 
      integer nmax
      character(len=1),target:: u_c(8192),v_c(8192)
      type(smprec) u,v,w,ui,vi
      double precision rx
      parameter (nmax=8192,rx=256.d0)
!u    uses fht_dp
      integer j,mn,nn
      double precision cy,t,a(nmax),b(nmax)
      u%m => u_c(1)
      u%x = 8192
      v%m => v_c(1)
      v%x = 8192
      u = ui
      v = vi
      w%l = u%l + v%l
      w%s = u%s*v%s
      mn=max(v%l,u%l)
      nn=1
      do while (nn.lt.mn)
       nn=nn+nn
      enddo
      nn=nn+nn
      if(nn.gt.nmax)then
       write(0,*)'nmax too small in fftmul'
       call exit(1)
      endif
      do 11 j=1,u%l
        a(j)=ichar(u%m(j:j))
11    continue
      do 12 j=u%l+1,nn
        a(j)=0.d0
12    continue
      do 13 j=1,v%l
        b(j)=ichar(v%m(j:j))
13    continue
      do 14 j=v%l+1,nn
        b(j)=0.d0
14    continue
      call smp_trigs(nn/2)
      call fht_dp(a,smp_trig,nn)
      call fht_dp(b,smp_trig,nn)
      call cnvl_ht_dp(b,a,nn)
      call fht_dp(b,smp_trig,nn)
      cy=0.
      do 16 j=nn,1,-1
        t=b(j)+cy+0.1d0
        b(j)=mod(t,rx)
        cy=floor(t/rx)
16    continue
      if (cy.ge.rx)then
       write(0,*)'should not happen in fftmul'
       call exit(1)
      endif
      w%m(1:1)=char(int(cy))
      do 17 j=2,w%l
        w%m(j:j)=char(int(b(j-1)))
17    continue
      if( smp_strip )then
       call smpzstrip(w,0)
      endif
      return
      end subroutine smpmul

      subroutine smp2dfr(a,s)
      integer iaz,ls,i,is
      character(len=1),target:: t_c(8192)
      type(smprec) a,s,t
      character(len=1) tc
      parameter (iaz=48)
      t%m => t_c(1)
      t%x = 8192


      t = a
      call smpzstrip(t,0)

      if( t%l .eq. 0)then  !  there's nothing left (should never happen)
       s%m(1:1) = char(iaz)
       s%l = 1
       return
      endif

      ls = 0
      do while (  t%l .gt. 0 )
          call smpsdv(t,t,10,is)
          ls = ls + 1
          s%m(ls:ls) = char(is+iaz)
          if( ichar(t%m(1:1)) .eq. 0)then
           call smplshft(t,1)
          endif
      enddo

      if( t%s .eq. -1 )then
       ls = ls + 1
       s%m(ls:ls) = "-"
      endif

      s%l = ls
      do i = 1,s%l/2
       tc = s%m(i:i)
       s%m(i:i) = s%m(s%l+1-i:s%l+1-i)
       s%m(s%l+1-i:s%l+1-i) = tc
      enddo
       
      return
      end subroutine smp2dfr


      subroutine smpinv(u,v,lu)
!    lu specifies how many bytes you want in the inverse
      character(len=1),target:: rr_c(8192),s_c(8192)
      integer :: mf,nmax,j,icnt,i,mm,ncnt=200,lu
      type(smprec) v,u,rr,s
      real bi
      parameter (mf=4,bi=1./256.,nmax=8192)
!     uses smpmul,smpsneg
      real fu,fv
      rr%m => rr_c(1)
      rr%x = 8192
      s%m => s_c(1)
      s%x = 8192
      u%l = lu
      u%s = v%s
      if(max(u%l,v%l).gt.nmax)then
       write(0,*)'nmax too small in smpinv'
       call exit(1)
      endif
!      call display(v,'entering smpinv with v')
      mm=min(mf,v%l)
      fv=ichar(v%m(mm:mm))
      do 11 j=mm-1,1,-1
        fv=fv*bi+ichar(v%m(j:j))
11    continue
      fu=1./fv
      do 12 j=1,u%l
        i=int(fu)
        u%m(j:j)=char(i)
        fu=256.*(fu-i)
12    continue
      icnt = 0
1     icnt = icnt + 1
!      if( icnt .eq. 1 )then
!       call display(u,'initial u')
!      elseif( icnt .eq. ncnt )then
!       call display(u,'next to last u')
!      endif
        call smpmul(rr,u,v)
!      call display(rr,'u*v')
        rr%l = u%l + 1
        call smplshft(rr,1)
        s = rr
        call smpsneg(s)
        s%m(1:1)=char(ichar(s%m(1:1))-254)
!      call display(s,'s')
        call smpmul(rr,s,u)
        rr%l = u%l + 1
        call smplshft(rr,1)
        u = rr
!      if( icnt .eq. 100 )then
!       call display(u,'last u')
!      endif
        call smplshft(s,1)
        s%l = s%l - 1
!      call display(s,'checking s')
        if( s .mne. zero_smp  )then
         if( icnt .lt. 100)then
          go to 1
         else
          write(0,*)' smpinv didn''t converge'
          return
         endif
        endif
!      write(*,*)' converged with icnt ',icnt
!      call display(u,'last u')
!      write(*,*)' leaving smpinv '
      return
      end subroutine smpinv

      logical function mge_smprec(ai,bi)
      type(smprec),intent(in) ::  ai
      type(smprec),intent(in) ::  bi
      type(smprec) a,b
      integer k
      character(len=1),target::a_c(8192),b_c(8192)
      a%m=>a_c(1)
      a%x=8192
      b%m=>b_c(1)
      b%x=8192
      a = ai
      b = bi
      call smpzstrip(a,0)
      call smpzstrip(b,0)
!      call display(a,'mge a')
!      call display(b,'mge b')

      if( a%l .gt. b%l )then
!       write(*,*)'a%l > b%l'
       mge_smprec = .true.
      elseif( a%l .lt. b%l )then
!       write(*,*)'a%l < b%l'
       mge_smprec = .false.
      else
       k = 1
       do while ( k .le. a%l )
        if( ichar(a%m(k:k)) .lt. ichar(b%m(k:k)) )then
         mge_smprec = .false.
!         write(*,*)'k,mge ',k,mge_smprec
         return
        elseif( ichar(a%m(k:k)) .gt. ichar(b%m(k:k)) )then
          mge_smprec = .true.
!          write(*,*)'k,mge ',k,mge_smprec
          return
        endif
       k = k + 1
       enddo
       mge_smprec = .true.
!       write(*,*)'= mge ',mge_smprec
      endif
!      write(*,*)'return mge ',mge_smprec
      return
      end function mge_smprec
        
      logical function meq_smprec(ai,bi)
      type(smprec),intent(in) ::  ai
      type(smprec),intent(in) ::  bi
      type(smprec) a,b
      integer k
      character(len=1),target::a_c(8192),b_c(8192)
      a%m=>a_c(1)
      a%x=8192
      b%m=>b_c(1)
      b%x=8192
      a = ai
      b = bi
      call smpzstrip(a,0)
      call smpzstrip(b,0)
      if( a%l .ne. b%l )then
       meq_smprec = .false.
      else
       k = 1
       do while ( k .le. a%l )
        if( ichar(a%m(k:k)) .ne. ichar(b%m(k:k)) )then
         meq_smprec = .false.
         return
        endif
       k = k + 1
       enddo
       meq_smprec = .true.
      endif
      return
      end function meq_smprec

      logical function eq_smprec(ai,bi)
      type(smprec),intent(in) ::  ai
      type(smprec),intent(in) ::  bi
      type(smprec) a,b
      integer k
      character(len=1),target::a_c(8192),b_c(8192)
      a%m=>a_c(1)
      a%x=8192
      b%m=>b_c(1)
      b%x=8192
      a = ai
      b = bi
      call smpzstrip(a,0)
      call smpzstrip(b,0)
      if( a%s .ne. b%s .or.  a%l .ne. b%l )then
       eq_smprec = .false.
      else
       k = 1
       do while ( k .le. a%l )
        if( ichar(a%m(k:k)) .ne. ichar(b%m(k:k)) )then
         eq_smprec = .false.
         return
        endif
       k = k + 1
       enddo
       eq_smprec = .true.
      endif
      return
      end function eq_smprec

      logical function ne_smprec(ai,bi)
      type(smprec),intent(in) ::  ai
      type(smprec),intent(in) ::  bi
      type(smprec) a,b
      integer k
      character(len=1),target::a_c(8192),b_c(8192)
      a%m=>a_c(1)
      a%x=8192
      b%m=>b_c(1)
      b%x=8192
      a = ai
      b = bi
      call smpzstrip(a,0)
      call smpzstrip(b,0)
      if( a%s .ne. b%s .or.  a%l .ne. b%l )then
       ne_smprec = .true.
      else
       k = 1
       do while ( k .le. a%l )
        if( ichar(a%m(k:k)) .ne. ichar(b%m(k:k)) )then
         ne_smprec = .true.
         return
        endif
       k = k + 1
       enddo
       ne_smprec = .false.
      endif
      return
      end function ne_smprec

      logical function mne_smprec(ai,bi)
      type(smprec),intent(in) ::  ai
      type(smprec),intent(in) ::  bi
      type(smprec) a,b
      integer k
      character(len=1),target::a_c(8192),b_c(8192)
      a%m=>a_c(1)
      a%x=8192
      b%m=>b_c(1)
      b%x=8192
      a = ai
      b = bi
      call smpzstrip(a,0)
      call smpzstrip(b,0)
      if( a%l .ne. b%l )then
       mne_smprec = .true.
      else
       k = 1
       do while ( k .le. a%l )
        if( ichar(a%m(k:k)) .ne. ichar(b%m(k:k)) )then
         mne_smprec = .true.
         return
        endif
       k = k + 1
       enddo
       mne_smprec = .false.
      endif
      return
      end function mne_smprec

      type(smprec) function plus_smprec(ui,vi) result(w)
      integer loc1,j
      type(smprec),intent(in) :: ui
      type(smprec),intent(in) :: vi
      type(smprec) :: u,v
      character*1 creg(4)
      character*4 cireg
      equivalence (cireg,creg)
      character(len=1),target::u_c(8192),v_c(8192)
      u%m=>u_c(1)
      u%x=8192
      v%m=>v_c(1)
      v%x=8192
!      call display(ui,'ui')
!      call display(vi,'vi')
      w%m=>plusfcn_c(1)
      w%x=8192
       cireg=transfer(0,cireg)
!       write(*,*)'initial cireg ',(ichar(creg(i)),i=1,4)
       u = ui
       v = vi
!      write(*,*)'u%l,v%l',u%l,v%l
!      call display(u,'u')
!      call display(v,'v')
       loc1 = max(u%l,v%l)   !  line up the units place
!       write(*,*)'u%l,v%l,loc1',u%l,v%l,loc1
       call smprshft(v,loc1-v%l)  
!       call display(v,'v after rshft')
       call smprshft(u,loc1-u%l)
!       call display(u,'u after rshft')

       if ( u%s .eq. v%s )then  !  u + , v +    .or. u - , v - 
!         call display(u,'before mge u')
!         call display(v,'before mge v')
        if( u .mge. v )then
         do  j=u%l,1,-1
!           write(*,*)'cireg j',j,(ichar(creg(i)),i=1,4)
           cireg=transfer(ichar(u%m(j:j))+ichar(v%m(j:j))+ichar(creg(smparith_ii2)),cireg)
           w%m(j+1:j+1)=creg(smparith_ii1)
!           write(*,*)'j+1 w%m(j+1)',j+1,ichar(w%m(j+1:j+1))
         enddo
         w%m(1:1)=creg(smparith_ii2)
!         write(*,*)'w%m(1)',(ichar(w%m(j:j)),j=1,w%l)
         w%s = u%s
         w%l = u%l + 1
        else  !  u .mlt. v
         do  j=v%l,1,-1
           cireg=transfer(ichar(v%m(j:j))+ichar(u%m(j:j))+ichar(creg(smparith_ii2)),cireg)
           w%m(j+1:j+1)=creg(smparith_ii1)
         enddo
         w%m(1:1)=creg(smparith_ii2)
         w%s = v%s
         w%l = v%l + 1
        endif
       elseif( u%s .eq. 1 )then  ! u + , v -
        if( u .mge. v )then
         cireg=transfer(256,cireg) 
         do  j=u%l,1,-1
           cireg=transfer(255+ichar(u%m(j:j))-ichar(v%m(j:j))+ichar(creg(smparith_ii2)),cireg)
           w%m(j:j)=creg(smparith_ii1)
         enddo
         w%s = u%s
         w%l = u%l
        else !  v .mgt. u 
         cireg=transfer(256,cireg) 
         do  j=v%l,1,-1
           cireg=transfer(255+ichar(v%m(j:j))-ichar(u%m(j:j))+ichar(creg(smparith_ii2)),cireg)
           w%m(j:j)=creg(smparith_ii1)
         enddo
         w%s = v%s
         w%l = v%l
        endif
       else  ! u - , v +
        if( v .mge. u )then
         cireg=transfer(256,cireg) 
         do  j=v%l,1,-1
           cireg=transfer(255+ichar(v%m(j:j))-ichar(u%m(j:j))+ichar(creg(smparith_ii2)),cireg)
           w%m(j:j)=creg(smparith_ii1)
         enddo
         w%s = v%s
         w%l = v%l
        else ! u .mgt. v
         cireg=transfer(256,cireg) 
         do  j=u%l,1,-1
           cireg=transfer(255+ichar(u%m(j:j))-ichar(v%m(j:j))+ichar(creg(smparith_ii2)),cireg)
           w%m(j:j)=creg(smparith_ii1)
         enddo
         w%s = u%s
         w%l = u%l
        endif
       endif  !  end u%s,v%s branches
!      call display(w,'before strip')
      if( smp_strip )then
       call smpzstrip(w,0)
      endif
!      call display(w,'after strip')
      return
      end function plus_smprec


      type(smprec) function sbtr_smprec(u,vi) result(w)
      type(smprec), intent(in) :: u,vi
      type(smprec) :: v
      character*1 creg(4)
      character*4 cireg
      equivalence (cireg,creg)
      character(len=1),target::v_c(8192)
      v%m=>v_c(1)
      v%x=8192
      w%m=>sbtrfcn_c(1)
      w%x=8192
      v = vi
      v%s = -v%s  !  flip the sign of v
      w = u + v
      return
      end function sbtr_smprec

      type(smprec) function mult_smprec(ui,vi) result(w)
      integer nmax
      type(smprec),intent(in):: ui,vi
      type(smprec) :: u,v
      double precision rx
      parameter (nmax=8192,rx=256.d0)
!u    uses fht_dp
      integer j,mn,nn
      double precision cy,t,a(nmax),b(nmax)
      character(len=1),target::u_c(8192),v_c(8192)
      u%m=>u_c(1)
      u%x=8192
      v%m=>v_c(1)
      v%x=8192
      w%m=>multfcn_c(1)
      w%x=8192
      u = ui
      v = vi
      w%l = u%l + v%l
      w%s = u%s*v%s
      mn=max(v%l,u%l)
      nn=1
      do while (nn.lt.mn)
       nn=nn+nn
      enddo
      nn=nn+nn
      if(nn.gt.nmax)then
       write(0,*)'nmax too small in fftmul'
       call exit(1)
      endif
      do 11 j=1,u%l
        a(j)=ichar(u%m(j:j))
11    continue
      do 12 j=u%l+1,nn
        a(j)=0.d0
12    continue
      do 13 j=1,v%l
        b(j)=ichar(v%m(j:j))
13    continue
      do 14 j=v%l+1,nn
        b(j)=0.d0
14    continue
      call smp_trigs(nn/2)
      call fht_dp(a,smp_trig,nn)
      call fht_dp(b,smp_trig,nn)
      call cnvl_ht_dp(b,a,nn)
      call fht_dp(b,smp_trig,nn)
      cy=0.
      do 16 j=nn,1,-1
        t=b(j)+cy+0.1d0
        b(j)=mod(t,rx)
        cy=floor(t/rx)
16    continue
      if (cy.ge.rx) then
       write(0,*)'should not happen in fftmul'
       call exit(1)
      endif
      w%m(1:1)=char(int(cy))
      do 17 j=2,w%l
        w%m(j:j)=char(int(b(j-1)))
17    continue
      if( smp_strip )then
       call smpzstrip(w,0)
      endif
      return
      end function mult_smprec

      type(smprec) function dvde_smprec(ui,vi) result(q)
      integer nmax,i,is
      type(smprec), intent(in) :: ui,vi
      type(smprec)  u,v,r,rr,s
      parameter (nmax=8192)
!u    uses smpinv,smpmul,smpssub
      character(len=1),target:: u_c(8192),v_c(8192),rr_c(8192),s_c(8192),&
      & r_c(8192)
      u%m => u_c(1)
      u%x = 8192
      v%m => v_c(1)
      v%x = 8192
      rr%m => rr_c(1)
      rr%x = 8192
      s%m => s_c(1)
      s%x = 8192
      q%m => dvdefcn_c(1)
      q%x = 8192
      r%m => r_c(1)
      r%x = 8192

1001  if( vi .meq. one_smp )then
       q = ui
       r = zero_smp
       if( vi%s < 0 )then
        q%s = -q%s
       endif
       return
      elseif( vi .meq. ui )then
       q = one_smp
       r = zero_smp
       if( vi%s /= ui%s )then
        q%s = -q%s
       endif
       return
      elseif( vi .mgt. ui )then
       q = zero_smp
       r = ui
       return
      elseif( vi .meq. zero_smp )then
       write(0,'(1x,a)')'error in dvde_smprec: v = 0'
       call exit(1)
      endif

      v = vi
      call smpzstrip(v,0)  !  used to be (v,1)
      smp_strip = .false.

      u = ui
      r%l = v%l
      q%l = u%l - v%l + 1
      if(u%l+smparith_macc.gt.nmax)then
        write(0,*)'u ',(ichar(u%m(i:i)),i=1,u%l)
        write(0,*)'v ',(ichar(v%m(i:i)),i=1,v%l)
        write(0,*)'nmax too small in dvde_smprec'
        call exit(1)
      endif
!      call display(u,'entering dvde_smprec with u=')
!      write(*,*)'u%l',u%l
!      write(*,*)(ichar(u%m(i:i)),i=1,u%l)
!      call display(v,'and v=')
!      write(*,*)'v%l',v%l
!      write(*,*)(ichar(v%m(i:i)),i=1,v%l)
      call smpinv(s,v,u%l-v%l+smparith_macc)
!      call display(s,'vinv')
!      write(*,*)'s%l',s%l
!      write(*,*)(ichar(s%m(i:i)),i=1,s%l)
      call smpmul(rr,s,u)
      rr%l = u%l - v%l + 2
      call smplshft(rr,1)
!      call display(rr,'setting q =')
      q = rr
      call smpmul(rr,q,v)
!      call display(rr,'q*v')
!      write(*,*)'q*v rr%l',rr%l
!      write(*,*)(ichar(rr%m(i:i)),i=1,rr%l)
      call smplshft(rr,1)
!      call display(rr,'lsfht 1 q*v = rr =')
!      write(*,*)'lshft 1 rr%l',rr%l
!      write(*,*)(ichar(rr%m(i:i)),i=1,rr%l)
      call smpssub(is,rr,u,rr)
      if(is .ne. 0)then
       smparith_macc = smparith_macc + 11
       if( smparith_macc > 4000 )then
        write(0,*)'smparith_macc > 4000 in dvde_smprec'
        call exit(1)
       endif
       go to 1001
      endif
      r = rr
!      call display(r,'r after r=rr')
!      write(*,*)'r%l',r%l
!      write(*,*)(ichar(r%m(i:i)),i=1,r%l)
      call smplshft(r,u%l-v%l)
!      call display(r,'lsfht u%l-v%l r')
!      write(*,*)'r%l',r%l
!      write(*,*)(ichar(r%m(i:i)),i=1,r%l)
      if( r .mge. v )then   !  sometimes the inverse doesn't quite do it !!!!
!       write(*,*)'|r| too large; u%s,v%s',u%s,v%s
!       call display(q,'before q')
!       call display(r,'before r')
       if( u%s*v%s > 0 )then
         q = q + one_smp
         r = r - v
       else
         q = q - one_smp
         r = r + v
       endif
!       call display(q,'after q')
!       call display(r,'after r')
      endif
      q%s = u%s*v%s
      r%s = u%s
      smp_strip = .true.
      call smpzstrip(q,0)
      call smpzstrip(r,0)
      return
      end function dvde_smprec


      logical function gt_smprec(ai,bi)
      type(smprec),intent(in) ::  ai
      type(smprec),intent(in) ::  bi
      type(smprec) a,b
      integer k
      character(len=1),target::a_c(8192),b_c(8192)
      a%m=>a_c(1)
      a%x=8192
      b%m=>b_c(1)
      b%x=8192
      a = ai
      b = bi
      call smpzstrip(a,0)
      call smpzstrip(b,0)

    if( a%s .eq. 1 .and. b%s .eq. 1)then
      if( a%l .gt. b%l )then
       gt_smprec = .true.
      elseif( a%l .lt. b%l )then
       gt_smprec = .false.
      else
       k = 1
       do while ( k .le. a%l )
        if( ichar(a%m(k:k)) .gt. ichar(b%m(k:k)) )then
         gt_smprec = .true.
         return
        elseif( ichar(a%m(k:k)) .lt. ichar(b%m(k:k)) )then
          gt_smprec = .false.
          return
        endif
       k = k + 1
       enddo
       gt_smprec = .false.
      endif
    elseif( a%s .eq. -1 .and. b%s .eq. 1 )then
      gt_smprec = .false.
    elseif( a%s .eq. 1 .and. b%s .eq. -1 )then
      gt_smprec = .true.
    else
      if( a%l .gt. b%l )then
       gt_smprec = .false.
      elseif( a%l .lt. b%l )then
       gt_smprec = .true.
      else
       k = 1
       do while ( k .le. a%l )
        if( ichar(a%m(k:k)) .lt. ichar(b%m(k:k)) )then
         gt_smprec = .true.
         return
        elseif( ichar(a%m(k:k)) .gt. ichar(b%m(k:k)) )then
          gt_smprec = .false.
          return
        endif
       k = k + 1
       enddo
       gt_smprec = .false.
      endif
    endif
      return
      end function gt_smprec
        

      subroutine display(a,label)
      type(smprec) a,sn
      character(len=10000) oline
      character(len=*) label
      character(len=1),target::sn_c(8192)
      integer j,k,ll,iu
      sn%m=>sn_c(1)
      sn%x=8192
       call smp2dfr(a,sn)
       ll = len_trim(label)
       k = 0
       do j = 1,ll
        k = k + 1
        oline(k:k) = label(j:j)
       enddo
       if( k > 0 )then
        k = k + 1
        oline(k:k) = " "
       endif
       do j = 1,sn%l
        k = k + 1
        oline(k:k) = sn%m(j:j)
       enddo
       write(*,'(a)')oline(1:k)
!       do i = 1,k,70
!        iu = min(i+69,k)
!        write(*,'(a)')oline(i:iu)
!       enddo
      return
      end subroutine display


      subroutine smpsqrt(w,vi)
      integer nmax,mf,i,ir,j,vll,icnt
      type(smprec) w,v,vi,u,un,t,q,r
      real bi,fu,fv,bb
      character(len=1),target:: v_c(8192),u_c(8192),un_c(8192),t_c(8192),q_c(8192),r_c(8192)
      parameter (nmax=8192,mf=3,bb=256.,bi=1./256.)
      v%m=>v_c(1)
      v%x=nmax
      u%m=>u_c(1)
      u%x=nmax
      un%m=>un_c(1)
      un%x=nmax
      t%m=>t_c(1)
      t%x=nmax
      q%m=>q_c(1)
      q%x=nmax
      r%m=>r_c(1)
      r%x=nmax
      if(w%x < (vi%l+1)/2)then
       write(0,*)'w%x too small in smpsqrt'
       call exit(1)
      endif
      v = vi
      vll = min(15,v%l)  !  can only handle 15 digits with real arithmetic
!      write(*,*)'vll = ',vll
!      call display(v,'v =')
      fv=ichar(v%m(1:1))
!      write(*,*)'initial fv',fv
      do 11 j=2,vll
        fv=bb*fv+ichar(v%m(j:j))
!        write(*,*)'j fv ',j,fv
11    continue
!      write(*,*)'final fv = ',fv
      fu=sqrt(fv)
!      write(*,*)'fu = ',fu
      do 12 j=vll,1,-1
        i=mod(fu,bb)
        u%m(j:j)=char(i)
        fu=int(fu/bb)
12    continue
      do i = vll+1,v%l  !  adjust for the missing digits here
       u%m(i:i) = char(0)
      enddo
      u%l = vi%l
      u%s = vi%s
      icnt = 0
1     icnt = icnt + 1
      call smpzstrip(u,0)
!      call display(u,'u =')
      call smpdiv(q,r,v,u)
!      call display(q,'v/u =')
!      call display(r,'r =')
      smp_strip = .false.
      call smpadd(t,q,u)
!      call display(t,'u + v/u =')
      call smpsdv(un,t,2,ir)
!      call display(un,'un = ')
      call smpsub(t,u,un)
!      call display(t,' old u - new u =')
      if(  t .mne. zero_smp )then
       if( t .ne. one_smp )then
        if( icnt .lt. 2000)then
         u = un
         goto 1
        else
         write(0,*)'smpsqrt didn''t converge'
         w = un
         smp_strip = .true.
         call smpzstrip(w,0)
         return
        endif
       endif
      endif
      w = un
      smp_strip = .true.
      call smpzstrip(w,0)
      return
      end subroutine smpsqrt


      logical function lt_smprec(ai,bi)
      type(smprec),intent(in) ::  ai
      type(smprec),intent(in) ::  bi
      type(smprec) a,b
      integer k
      character(len=1),target::a_c(8192),b_c(8192)
      a%m=>a_c(1)
      a%x=8192
      b%m=>b_c(1)
      b%x=8192
      a = ai
      b = bi
      call smpzstrip(a,0)
      call smpzstrip(b,0)

    if( a%s .eq. 1 .and. b%s .eq. 1)then
      if( a%l .lt. b%l )then
       lt_smprec = .true.
      elseif( a%l .gt. b%l )then
       lt_smprec = .false.
      else
       k = 1
       do while ( k .le. a%l )
        if( ichar(a%m(k:k)) .lt. ichar(b%m(k:k)) )then
         lt_smprec = .true.
         return
        elseif( ichar(a%m(k:k)) .gt. ichar(b%m(k:k)) )then
          lt_smprec = .false.
          return
        endif
       k = k + 1
       enddo
       lt_smprec = .false.
      endif
    elseif( a%s .eq. -1 .and. b%s .eq. 1 )then
      lt_smprec = .true.
    elseif( a%s .eq. 1 .and. b%s .eq. -1 )then
      lt_smprec = .false.
    else
      if( a%l .gt. b%l )then
       lt_smprec = .true.
      elseif( a%l .lt. b%l )then
       lt_smprec = .false.
      else
       k = 1
       do while ( k .le. a%l )
        if( ichar(a%m(k:k)) .lt. ichar(b%m(k:k)) )then
         lt_smprec = .false.
         return
        elseif( ichar(a%m(k:k)) .gt. ichar(b%m(k:k)) )then
          lt_smprec = .true.
          return
        endif
       k = k + 1
       enddo
       lt_smprec = .false.
      endif
    endif
      return
      end function lt_smprec

      logical function ge_smprec(ai,bi)
      type(smprec),intent(in) ::  ai
      type(smprec),intent(in) ::  bi
      type(smprec) a,b
      integer k
      character(len=1),target::a_c(8192),b_c(8192)
      a%m=>a_c(1)
      a%x=8192
      b%m=>b_c(1)
      b%x=8192
      a = ai
      b = bi
      call smpzstrip(a,0)
      call smpzstrip(b,0)
    if( a%s .eq. 1 .and. b%s .eq. 1)then
      if( a%l .gt. b%l )then
       ge_smprec = .true.
      elseif( a%l .lt. b%l )then
       ge_smprec = .false.
      else
       k = 1
       do while ( k .le. a%l )
        if( ichar(a%m(k:k)) .gt. ichar(b%m(k:k)) )then
         ge_smprec = .true.
         return
        elseif( ichar(a%m(k:k)) .lt. ichar(b%m(k:k)) )then
          ge_smprec = .false.
          return
        endif
       k = k + 1
       enddo
       ge_smprec = .true.
      endif
    elseif( a%s .eq. -1 .and. b%s .eq. 1 )then
      ge_smprec = .false.
    elseif( a%s .eq. 1 .and. b%s .eq. -1 )then
      ge_smprec = .true.
    else
      if( a%l .gt. b%l )then
       ge_smprec = .false.
      elseif( a%l .lt. b%l )then
       ge_smprec = .true.
      else
       k = 1
       do while ( k .le. a%l )
        if( ichar(a%m(k:k)) .lt. ichar(b%m(k:k)) )then
         ge_smprec = .true.
         return
        elseif( ichar(a%m(k:k)) .gt. ichar(b%m(k:k)) )then
          ge_smprec = .false.
          return
        endif
       k = k + 1
       enddo
       ge_smprec = .true.
      endif
    endif
      return
      end function ge_smprec

      logical function le_smprec(ai,bi)
      type(smprec),intent(in) ::  ai
      type(smprec),intent(in) ::  bi
      type(smprec) a,b
      integer k
      character(len=1),target::a_c(8192),b_c(8192)
      a%m=>a_c(1)
      a%x=8192
      b%m=>b_c(1)
      b%x=8192
      a = ai
      b = bi
      call smpzstrip(a,0)
      call smpzstrip(b,0)
    if( a%s .eq. 1 .and. b%s .eq. 1)then
      if( a%l .lt. b%l )then
       le_smprec = .true.
      elseif( a%l .gt. b%l )then
       le_smprec = .false.
      else
       k = 1
       do while ( k .le. a%l )
        if( ichar(a%m(k:k)) .lt. ichar(b%m(k:k)) )then
         le_smprec = .true.
         return
        elseif( ichar(a%m(k:k)) .gt. ichar(b%m(k:k)) )then
          le_smprec = .false.
          return
        endif
       k = k + 1
       enddo
       le_smprec = .true.
      endif
    elseif( a%s .eq. -1 .and. b%s .eq. 1 )then
      le_smprec = .true.
    elseif( a%s .eq. 1 .and. b%s .eq. -1 )then
      le_smprec = .false.
    else
      if( a%l .gt. b%l )then
       le_smprec = .true.
      elseif( a%l .lt. b%l )then
       le_smprec = .false.
      else
       k = 1
       do while ( k .le. a%l )
        if( ichar(a%m(k:k)) .lt. ichar(b%m(k:k)) )then
         le_smprec = .false.
         return
        elseif( ichar(a%m(k:k)) .gt. ichar(b%m(k:k)) )then
          le_smprec = .true.
          return
        endif
       k = k + 1
       enddo
       le_smprec = .true.
      endif
    endif
      return
      end function le_smprec

      logical function mgt_smprec(ai,bi)
      type(smprec),intent(in) ::  ai
      type(smprec),intent(in) ::  bi
      type(smprec) a,b
      integer k
      character(len=1),target::a_c(8192),b_c(8192)
      a%m=>a_c(1)
      a%x=8192
      b%m=>b_c(1)
      b%x=8192
      a = ai
      b = bi
      call smpzstrip(a,0)
      call smpzstrip(b,0)
      if( a%l .gt. b%l )then
       mgt_smprec = .true.
      elseif( a%l .lt. b%l )then
       mgt_smprec = .false.
      else
       k = 1
       do while ( k .le. a%l )
        if( ichar(a%m(k:k)) .lt. ichar(b%m(k:k)) )then
         mgt_smprec = .false.
         return
        elseif( ichar(a%m(k:k)) .gt. ichar(b%m(k:k)) )then
          mgt_smprec = .true.
          return
        endif
       k = k + 1
       enddo
       mgt_smprec = .false.
      endif
      return
      end function mgt_smprec

      logical function mlt_smprec(ai,bi)
      type(smprec),intent(in) ::  ai
      type(smprec),intent(in) ::  bi
      type(smprec) a,b
      integer k
      character(len=1),target::a_c(8192),b_c(8192)
      a%m=>a_c(1)
      a%x=8192
      b%m=>b_c(1)
      b%x=8192
      a = ai
      b = bi
      call smpzstrip(a,0)
      call smpzstrip(b,0)
      if( a%l .gt. b%l )then
       mlt_smprec = .false.
      elseif( a%l .lt. b%l )then
       mlt_smprec = .true.
      else
       k = 1
       do while ( k .le. a%l )
        if( ichar(a%m(k:k)) .lt. ichar(b%m(k:k)) )then
         mlt_smprec = .true.
         return
        elseif( ichar(a%m(k:k)) .gt. ichar(b%m(k:k)) )then
          mlt_smprec = .false.
          return
        endif
       k = k + 1
       enddo
       mlt_smprec = .false.
      endif
      return
      end function mlt_smprec

      logical function mle_smprec(ai,bi)
      type(smprec),intent(in) ::  ai
      type(smprec),intent(in) ::  bi
      type(smprec) a,b
      integer k
      character(len=1),target::a_c(8192),b_c(8192)
      a%m=>a_c(1)
      a%x=8192
      b%m=>b_c(1)
      b%x=8192
      a = ai
      b = bi
      call smpzstrip(a,0)
      call smpzstrip(b,0)
      if( a%l .gt. b%l )then
       mle_smprec = .false.
      elseif( a%l .lt. b%l )then
       mle_smprec = .true.
      else
       k = 1
       do while ( k .le. a%l )
        if( ichar(a%m(k:k)) .lt. ichar(b%m(k:k)) )then
         mle_smprec = .true.
         return
        elseif( ichar(a%m(k:k)) .gt. ichar(b%m(k:k)) )then
          mle_smprec = .false.
          return
        endif
       k = k + 1
       enddo
       mle_smprec = .true.
      endif
      return
      end function mle_smprec

      subroutine inum(u,cstrng,lstrng)
!
!  this subroutine will convert an input decimal string into
!  a type (smprec) number
!
      type(smprec) u,ten_smp,mx,digit
      character(len=1),target:: ten_c(1),mx_c(8192),digit_c(1)
      integer lstrng,ibgn,i
      character(len=*)cstrng
!      character(len=13)label

!  inum assumes that smpinit has been called
!  lstrng is the length of the input string cstrng

      ten_smp%s = 1
      ten_smp%l = 1
      ten_smp%m => ten_c(1)
      ten_smp%x = 1
      ten_smp%m(1:1) = char(10)

      mx%x = 8192
      mx%m => mx_c(1)

      digit%x = 1
      digit%m => digit_c(1)


      mx = one_smp


      if(cstrng(1:1) == "-")then
       u%s = -1
       mx%s = -1
       ibgn = 2
      elseif( cstrng(1:1) == "+")then
       u%s = +1
       ibgn = 2
      else
       u%s = +1
       ibgn = 1
      endif


      u%l = 1
      u%m(1:1) = char( ichar(cstrng(lstrng:lstrng)) - 48 )
      digit%s = 1
      digit%l = 1
      do i=lstrng-1,ibgn,-1
       mx = ten_smp*mx
       digit%m(1:1) = char( ichar(cstrng(i:i)) - 48 )
       u = u + mx*digit
      enddo


      return

      end subroutine inum

      subroutine i2smp(u,iiu)
      type(smprec) u,ten_smp,mx,digit
      character(len=1),target :: ten_smp_c,digit_c,mx_c(20)
      integer iu,iiu,i
!      character(len=30)label

      ten_smp%m=>ten_smp_c
      ten_smp%x=1
      ten_smp%s=1
      ten_smp%l=1
      ten_smp%m(1:1) = char(10)

      mx%m=>mx_c(1)
      mx%x=20

      digit%m=>digit_c
      digit%x=1

      mx = one_smp
      iu = abs(iiu)

      if( iiu < 0 )then
       u%s = -1
      else
       u%s = 1
      endif



      u%l = 1
      u%m(1:1) = char( mod(iu,10) )
!      call display(u,'initial u')
      digit%s = 1
      digit%l = 1
      i = 2
      do while ( iu > 0 )
       mx = ten_smp*mx
       iu = iu/10
       digit%m(1:1) = char( mod(iu,10) )
!       write(label,'(''digit'',i4)')i
!       call display(digit,label)
       u = u + mx*digit
!       call display(u,'next u')
       i = i + 1
      enddo


      return

      end subroutine i2smp

      subroutine smp2i(iu,a)

      type(smprec) a,t
      character(len=1),target::t_c(8192)
      integer mx,iu,is

      t%m=>t_c(1)
      t%x=8192



      t = a
      call smpzstrip(t,0)

      if( t%l .eq. 0)then  !  there's nothing left (should never happen)
       iu = 0
       return
      endif

      mx = 1
      iu = 0
      do while (  t%l .gt. 0 )
          call smpsdv(t,t,10,is)
          iu = iu + mx*is
          if( ichar(t%m(1:1)) .eq. 0 )then
           call smplshft(t,1)
          endif
          mx = 10*mx
      enddo

      iu = t%s*iu
       
      return
      end subroutine smp2i




      end module smparith
